# -*- coding: utf-8 -*-
import json


class Subscriber(object):
    """
    Chariot gateway subscriber
    """
    def __init__(self, subscriber_id):
        self.id = subscriber_id
        self.sensors = set()

    def dict(self):
        return {
            'id': self.id,
            'sensors': list(self.sensors)
        }

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return json.dumps(self.dict())



# -*- coding: utf-8 -*-

import os
import json
import logging
import logging.config


default_logging = {
    "version": 1,
    "handlers": {            
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "console"
        }
    },
    "loggers": {            
    },
    "root": {
        "level": "INFO",
        "handlers": [
            "console"
        ]
    },
    "formatters": {
        "console": {
            "format": "%(asctime)s (%(levelname)s) %(name)s : %(message)s"
        }
    }
}


class Brokers(object):
    def __init__(self, options):
        self.southbound = options.get('southbound', {})
        self.northbound = options.get('northbound', {})

class Configuration(object):
    def __init__(self, options):
        self.cloudant = options.get('cloudant', { "enabled": False })
        self.watson_iot =  options.get('watson_iot', { "enabled": False })
        self.local_storage =  options.get('local_storage', { "enabled": False })
        self.brokers =  Brokers(options.get('brokers', {}))
        self.tracer =  options.get('tracer', { "enabled": False })
        self.privacy_engine =  options.get('privacy_engine', None)
        self.alert_digester =  options.get('alert_digester', None)
        self.dispatcher =  options.get('dispatcher', None)
        self.northbound_dispatcher =  options.get('northbound_dispatcher', None)
        self.topology =  options.get('topology', None)
        self.set_logging( options.get('logging', default_logging))

    def set_logging(self, options):
        self.logging = options
        logging.config.dictConfig(self.logging)


def open_config_file(files = ['./config.json', './tests/config.json']):
    filename = None
    for name in files:
        if os.path.isfile(name):
            filename = name

    if filename is None:
        raise Exception('Configuration file is not exists')

    with open(filename, 'r') as read_file:
        opts = json.load(read_file)

    return Configuration(opts)


# -*- coding: utf-8 -*-
from influxdb import InfluxDBClient


class LocalDataSource(object):

    def __init__(self, hostname, port, username, password, db_name, duration='4w'):
        self.db = InfluxDBClient(hostname, port, username, password, db_name)
        self.db.create_database(db_name)
        self.db.create_retention_policy('awesome_policy', duration, 3, default=True)

    def publish(self, point):
        json_body = [
            {
                'measurement': point.table,
                'tags': {
                    'topic': point.topic,
                    'sensor_id': point.sensor_id
                },
                'time': point.timestamp,
                'fields': point.message
            }
        ]
        return self.db.write_points(json_body, protocol='json', retention_policy='awesome_policy')

    def query(self, q, db_name):
        return self.db.query(q, database=db_name)



# -*- coding: utf-8 -*-
import json
import datetime


class Alert(object):
    """
    Describe each alert raised by Chariot
    """
    def __init__(self, name, msg=None, severity=100):
        self.timestamp = datetime.datetime.utcnow().isoformat()
        self.name = name
        self.message = msg
        self.severity = severity
        self.sensor_id = None

    def dict(self):
        """
        Converts alert object to dictionary
        """
        return {
            'name': self.name,
            'sensor_id': self.sensor_id,
            'message': self.message,
            'severity': self.severity,
            'timestamp': self.timestamp
        }

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return json.dumps(self.dict())



from jaeger_client import Config
from opentracing import Format


class Tracer(object):
    def __init__(self, options):
        self.name = options['service']
        self.host = options['host']
        self.port = options['port']
        self.tracer = None

    def init_tracer(self):
        config = Config(
            config={
                'sampler': {
                    'type': 'const',
                    'param': 1,
                },
                'local_agent': {
                    'reporting_host': self.host,
                    'reporting_port': self.port
                },
                'logging': False,
            },            
            service_name=self.name,
        )

        self.tracer = config.initialize_tracer()

    def inject(self, span, msg):
        d = {}
        self.tracer.inject(span.context, Format.TEXT_MAP, d)
        msg['uber-trace-id'] = d['uber-trace-id']
        return msg

    def extract(self, msg):
        d = {
            'uber-trace-id': msg['uber-trace-id']
        }
        return self.tracer.extract(Format.TEXT_MAP, d)

    def close(self):
        self.tracer.close()


class Traceable:
    def __init__(self):
        self.tracer

    def inject_tracer(self, tracer):
        self.tracer = tracer

    def set_up_tracer(self, options):
        self.tracer = Tracer(options)
        self.tracer.init_tracer()

    def start_span(self, id, child_span=None):
        if self.tracer is None:
            return

        if child_span is None:
            return self.tracer.tracer.start_span(id)
        else:
            return self.tracer.tracer.start_span(id, child_of=child_span)

    def close_span(self, span):
        if self.tracer is None:
            return
        span.finish()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import uuid
import gmqtt
from ..utilities import Tracer


class LocalConnector:
    """
    All subscribers/publisers at Chariot project should extend this class
    """
    def __init__(self):
        self.connack = None
        self.client = None
        self.tracer = None

        self.disconnected = False
        self.connected = False

    def clear(self):
        """
        Reinitialize the connector
        """
        self.__init__()

    def subscribe(self, topic, qos=1):
        """
        Subscribe to a topic

        :param topic: Which topic the subscriber should listen for new message.
        :param qos: MQTT broker quality of service
        """
        self.client.subscribe(topic, qos=qos)

    def publish(self, topic, msg, qos=1):
        """
        Publish to a topic

        :param msg: the published message
        :param qos: MQTT broker quality of service
        """
        self.client.publish(topic, msg, qos=qos)

    def on_message(self, client, topic, payload, qos, properties):
        """
        Handler for new message

        :param client: the subscribed MQTT client
        :param topic: the MQTT topic
        :param payload: the message
        :param qos: MQTT broker quality of service
        :param properties: Custom properties
        """

    def on_subscribe(self, client, mid, qos):
        """
        The handler run when the client subscribed to a new topic

        :param client: the subscribed MQTT client
        :param mid:
        :param qos: MQTT broker quality of service
        """

    def on_disconnect(self, client, packet):
        """
        The handler run when the connections is fineshed

        :param client: the subscribed MQTT client
        :param packet:
        """
        self.disconnected = True

    def on_connect(self, client, flags, rc, properties):
        """
        The handler run when the connections is established

        :param client: the subscribed MQTT client
        :param flags:
        :param rc:
        :param properties: Custom properties
        """
        self.connected = True
        self.connack = (flags, rc, properties)

    def register_for_client(self, client):
        """
        Register handlers to the client.

        :param client: Client to register handlers
        """
        client.on_disconnect = self.on_disconnect
        client.on_message = self.on_message
        client.on_connect = self.on_connect
        client.on_subscribe = self.on_subscribe

        self.client = client

    def inject_tracer(self, tracer):
        """
        Inject an opentracing client

        :param tracer: the opentracing client
        """
        self.tracer = tracer

    def set_up_tracer(self, options):
        """
        Configure a new opentracing client

        :param options: options to configure the new client
        """
        self.tracer = Tracer(options)
        self.tracer.init_tracer()

    def start_span(self, span_id, child_span=None):
        """
        Start a new logging span

        :param span_id: identifier of a new logging span
        :param child_span: parent span
        """
        if self.tracer is None:
            return None

        if child_span is None:
            return self.tracer.tracer.start_span(span_id)
        
        return self.tracer.tracer.start_span(span_id, child_of=child_span)

    def close_span(self, span):
        """
        Close a logging span

        :param span: Span to close
        """
        if self.tracer is None:
            return
        span.finish()


async def create_client(options, postfix='_client'):
    """
    Create a new GMQTT client

    :param options: Options for client initialization
    :para postfix: Unique postfix for the client
    """
    client_id = '%s%s' % (uuid.uuid4(), postfix)

    client = gmqtt.Client(client_id, clean_session=True)
    await client.connect(host=options['host'], port=options['port'], version=4)

    return client



import math
import binascii
import base64
import hashlib
import hmac
import struct
import os
import textwrap

from ecdsa import numbertheory

from Crypto.Cipher import AES
from Crypto.Util import Counter
from Crypto.Random import random

from fastecdsa import curve, ecdsa, keys
from fastecdsa.point import Point


BASE_CURVE = curve.P256


def generate_new_keypair(curve=BASE_CURVE):
    priv_key, pub_key = keys.gen_keypair(curve)
    return KeyPair(pub_key, priv_key, curve)


def pad(message, block_size):
    """
    PKCS#7 padding
    """
    padded = message
    last_block = len(message) % block_size
    to_pad = block_size - last_block
    for i in range(to_pad):
        padded = padded + chr(to_pad)
    return padded


def unpad(message, block_size):
    if type(message) is bytes:
        return binary_unpad(message, block_size)

    length = len(message)
    if length == 0:
        return message

    to_pad = ord(message[length - 1])
    if to_pad > block_size:
        return message

    if length < to_pad:
        return message

    pad_start = length - to_pad
    for c in message[pad_start:]:
        if c != chr(to_pad):
            return message

    return message[:pad_start]


def binary_unpad(message, block_size):
    length = len(message)
    if length == 0:
        return message

    to_pad = message[length - 1]
    if to_pad > block_size:
        return message

    if length < to_pad:
        return message

    pad_start = length - to_pad
    return message[:pad_start]


def string_to_number(string):
    return int(binascii.hexlify(string), 16)


def number_to_string(num, order):
    length = orderlen(order)
    fmt_str = "%0" + str(2 * length) + "x"
    string = binascii.unhexlify((fmt_str % num).encode())
    assert len(string) == length, (len(string), length)
    return string


def orderlen(order):
    return (1 + len("%x" % order)) // 2  # bytes


def randrange(order, entropy=None):
    """Return a random integer k such that 1 <= k < order, uniformly
    distributed across that range. For simplicity, this only behaves well if
    'order' is fairly close (but below) a power of 256. The try-try-again
    algorithm we use takes longer and longer time (on average) to complete as
    'order' falls, rising to a maximum of avg=512 loops for the worst-case
    (256**k)+1 . All of the standard curves behave well. There is a cutoff at
    10k loops (which raises RuntimeError) to prevent an infinite loop when
    something is really broken like the entropy function not working.
    Note that this function is not declared to be forwards-compatible: we may
    change the behavior in future releases. The entropy= argument (which
    should get a callable that behaves like os.urandom) can be used to
    achieve stability within a given release (for repeatable unit tests), but
    should not be used as a long-term-compatible key generation algorithm.
    """

    if entropy is None:
        entropy = os.urandom
    assert order > 1
    bytes = orderlen(order)
    dont_try_forever = 10000  # gives about 2**-60 failures for worst case
    while dont_try_forever > 0:
        dont_try_forever -= 1
        candidate = string_to_number(entropy(bytes)) + 1
        if 1 <= candidate < order:
            return candidate
        continue
    raise RuntimeError("randrange() tried hard but gave up, either something"
                       " is very wrong or you got really unlucky. Order was"
                       " %x" % order)


def encode_point(p, compressed, curve=BASE_CURVE):
    order = curve.q
    x_str = number_to_string(p.x, order)
    if compressed:
        return bytes(chr(2 if (p.y & 1) == 0 else 3), 'utf-8') + x_str
    else:
        y_str = number_to_string(p.y, order)
    return bytes(chr(4), 'utf-8') + x_str + y_str


def decode_point(point, curve=BASE_CURVE):
    # See http://www.secg.org/download/aid-780/sec1-v2.pdf section 2.3.4
    elliptic_curve = curve
    order = elliptic_curve.q
    base_length = orderlen(order)

    if point[0] == 4:
        # 3
        x_str = point[1:base_length + 1]
        y_str = point[base_length + 1:]
        return Point(string_to_number(x_str), string_to_number(y_str), curve=elliptic_curve)
    else:
        # 2.3
        if point[0] == 2:
            yp = 0
        elif point[0] == 3:
            yp = 1
        else:
            return None
        # 2.2
        x_str = point[1:base_length + 1]
        x = string_to_number(x_str)
        # 2.4.1
        alpha = ((x * x * x) + (elliptic_curve.a * x) + elliptic_curve.b) % elliptic_curve.p
        beta = numbertheory.square_root_mod_prime(alpha, elliptic_curve.p)
        if (beta - yp) % 2 == 0:
            y = beta
        else:
            y = elliptic_curve.p - beta

    return Point(x, y, curve=elliptic_curve)


class KeyPair:
    def __init__(self, public_key, private_key, curve=BASE_CURVE):
        self.public_key = public_key
        self.private_key = private_key
        self.curve = curve

    def sign(self, message):
        r, s = ecdsa.sign(message, self.private_key)
        return self.encode_signature(r, s)

    def verify(self, message, r_s):
        r, s = self.decode_signature(r_s)
        return ecdsa.verify((r, s), message, self.public_key)

    def encode_signature(self, r, s):
        length = orderlen(self.curve.q)
        fmt_str = "%0" + str(2 * length) + "x"
        s = binascii.unhexlify((fmt_str % s).encode())
        r = binascii.unhexlify((fmt_str % r).encode())
        return base64.b64encode(r + s)

    def decode_signature(self, r_s):
        s_decoded = base64.b64decode(r_s)
        base_length = orderlen(self.curve.q)
        r = int(binascii.hexlify(s_decoded[:base_length]), 16)
        s = int(binascii.hexlify(s_decoded[base_length:]), 16)
        return r, s

    def decrypt(self, message, curve=BASE_CURVE):
        R_size = 1 + orderlen(curve.q)
        mac_size = hashlib.sha256().digest_size

        message_binary = base64.b64decode(message)
        if len(message_binary) < (R_size + mac_size):
            return None

        R = decode_point(message_binary)
        d = message_binary[R_size:R_size + mac_size]
        prefix_bytes = message_binary[R_size + mac_size:R_size + mac_size + 8]
        c = message_binary[R_size + mac_size + 8:]
        S = (self.private_key * R).x
        S_bytes = number_to_string(S, curve.q)
        k_E = hashlib.sha256(S_bytes + struct.pack(">I", 1)).digest()
        k_M = hashlib.sha256(S_bytes + struct.pack(">I", 2)).digest()
        d_verify = hmac.new(k_M, prefix_bytes + c, hashlib.sha256).digest()
        if d_verify != d:
            return None
        ctr = Counter.new(64, prefix=prefix_bytes)
        cipher = AES.new(key=k_E, mode=AES.MODE_CTR, counter=ctr)
        padded = cipher.decrypt(c)
        return unpad(padded, AES.block_size)

    def encrypt(self, message, curve=BASE_CURVE):
        padded = pad(message, AES.block_size).encode("utf-8")
        r = randrange(curve.q)
        R = curve.G * r
        S = (self.public_key * r).x
        S_bytes = number_to_string(S, curve.q)

        k_E = hashlib.sha256(S_bytes + struct.pack(">I", 1)).digest()
        k_M = hashlib.sha256(S_bytes + struct.pack(">I", 2)).digest()

        prefix = random.getrandbits(64)
        prefix_bytes = struct.pack("<Q", prefix)
        ctr = Counter.new(64, prefix=prefix_bytes)
        cipher = AES.new(key=k_E, mode=AES.MODE_CTR, counter=ctr)
        c = cipher.encrypt(padded)
        d = hmac.new(k_M, prefix_bytes + c, hashlib.sha256).digest()

        return textwrap.fill(base64.b64encode(encode_point(R, True) + d + prefix_bytes + c).decode(), 200)



# -*- coding: utf-8 -*-
import json
import uuid
import datetime
from ..utilities.parsing import try_parse, normalize_mac_address


FIXEDIO = 'fixedIO'
WIFI = 'wifi'
SENSORDATA = 'sensorData'
SENSORVALUES = 'sensorValues'
SENSORNAME = 'sensorName'

class DataPointFactory(object):
    """
    Converts to a new point
    """
    def __init__(self, db, table):
        self.db = db
        self.table = table

    def from_mqtt_message(self, message):
        """
        From mosquitto message payload
        """
        msg = message.payload.decode('utf-8')
        messages_parsed = self.from_json_string(msg)

        i = 0
        for message_parsed in messages_parsed:
            messages_parsed[i].topic = message.topic
            i = i + 1

        return messages_parsed

    def from_json_string(self, msg):
        """
        From JSON message payload
        """
        decoded_msg = json.loads(msg)
        messages = []
        for key, message in decoded_msg.items():
            key = normalize_mac_address(key)
            parsed_msg = None
            if FIXEDIO in message:
                parsed_msg = message[FIXEDIO]
                key = 'gateway_%s' % key
            elif WIFI in message:
                obj = {}
                for values in message[WIFI][SENSORDATA][SENSORVALUES]:
                    obj[values['name']] = try_parse(values['value'])
                parsed_msg = obj
                key = 'device_%s_%s' % (key, message[WIFI][SENSORDATA][SENSORNAME])
            else:
                raise Exception('Message format is not recognized')

            point = DataPoint(self.db, self.table, parsed_msg)
            point.sensor_id = key
            messages.append(point)
        return messages

class DataPoint(object):
    def __init__(self, db, table, message):
        self.id = uuid.uuid4()
        self.db = db
        self.table = table
        self.message = message
        self.timestamp = datetime.datetime.utcnow().isoformat()
        self.topic = None
        self.sensor_id = None

    def _event_type(self):
        if self.topic is None:
            return ''
        return self.topic.replace('/', '.')



from cloudant import Cloudant
from datetime import datetime


class CloudDataSource(object):

    def __init__(self, options):
        self.my_database = None
        self.orgId = options['orgId']
        self.cloudant_client = Cloudant.iam(options['username'], options['apikey'], connect=True)

    def get_last(self, device_type, device_id=None):
        """
        Gets the last reading of a sensor in a gateway from Cloudant.
        :param device_type: name of the gateway
        :param device_id: name of the sensor
        :return: timestamp and value in json format.
        """
        current_date = datetime.strftime(datetime.now(), '%Y-%m-%d')  # 2018-11-10
        self.my_database = self.cloudant_client['iotp_%s_default_%s' % (self.orgId, current_date)]

        if device_id is None:
            selector = dict(eventType='message',
                            deviceType={'$eq': device_type},
                            _id={'$gt': 0})
        else:
            selector = dict(deviceId=device_id,
                            deviceType={'$eq': device_type},
                            _id={'$gt': 0})

        result_collection = self.my_database.get_query_result(selector,
                                                              fields=['data'], sort=[{'_id': 'desc'}], limit=1)
        if len(result_collection[:]) > 0:
            return result_collection[0][0]['data']
        else:
            return None

    def get_all_database(self):
        return self.cloudant_client.all_dbs()



# -*- coding: utf-8 -*-
"""
Parsing string utilities
"""

import re


def try_parse(serialized_num):
        try:
            return int(serialized_num)
        except:
            try: 
                return float(serialized_num)
            except:
                return serialized_num


def normalize_mac_address(mac_address):
    regex = r"[:_-]"
    subst = ""
    return re.sub(regex, subst, mac_address, 0)


from multiprocessing import Process

import ibmiotf.gateway


class WatsonConnector(Process):

    def __init__(self, options):
        super(WatsonConnector, self).__init__()
        try:
            self.iot_client = ibmiotf.gateway.Client(options)
            self.iot_client.connect()
        except ibmiotf.ConnectionException as e:
            print(e)

    def publish(self, point):
        """
        :param point: A point represent received message.
        :return: True if event publish is successfull.
        """
        return self.iot_client.publishGatewayEvent(event=point.table, msgFormat="json", data=point.message)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('./README.rst') as readme_file:
    readme = readme_file.read()

with open('./HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'paho-mqtt',
    'asyncio',
    'gmqtt',
    'influxdb',
    'cloudant',
    'ibmiotf',
    'pytest',
    'fastecdsa',
    'ecdsa',
    'pycrypto',
    'jaeger-client',
    'pytest-asyncio', 
    # 'iotl',
]

# dependency_links = [
#     'git+ssh://git@gitlab.com/chariot-h2020/iot-modeling-language.git#egg=iotl-0.106'
# ]

setup_requirements = [ ]

test_requirements = [ ]

setup(
    author="George Theofilis",
    author_email='g.theofilis@clmsuk.com',
    classifiers=[
        'License :: OSI Approved :: Eclipse Public License 1.0 (EPL-1.0)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    description="Chariot base micro-service",
    install_requires=requirements,
    # dependency_links=dependency_links,
    license="EPL-1.0",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='chariot_base',
    name='chariot_base',
    packages=find_packages(include=[
        'chariot_base',
        'chariot_base.*'
    ]),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://gitlab.com/chariot-h2020/chariot_base',
    version='0.6.4',
    zip_safe=False,
)



# -*- coding: utf-8 -*-
from iotl import interpreter

class IoTLWrapper(object):
    
    def __init__(self, options):
        self.filepath = options['filepath']
        self.IoTState = interpreter.IoTState()

    def load(self):
        with open(self.filepath, 'r') as f:
            self.IoTState.parse(f.read())
        return True

    def isSensitive(self, sensor_id):
        try:
            return self.IoTState.params[sensor_id]['privacySensitive'] == 1
        except KeyError:
            return False

    def acl(self, sensor_id):
        try:
            return self.IoTState.acl[sensor_id]
        except KeyError:
            return []


# -*- coding: utf-8 -*-
import json


class Message(object):
    """
    Describe message passed between Chariot's components
    """
    def __init__(self, sensor_id, value):
        self.sensor_id = sensor_id
        self.value = value
        self.destination = None

    def dict(self):
        return {
            'sensor_id': self.sensor_id,
            'value': self.value,
            'destination': self.destination
        }

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return json.dumps(self.dict())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
import asyncio
import gmqtt
import logging

from chariot_base.connector import LocalConnector


class Callbacks(LocalConnector):

    def __init__(self):
        super().__init__()
        self.messages = []
        self.messagedicts = []
        self.unsubscribeds = []
        self.disconnected = []
        self.publisheds = []
        self.subscribeds = []
        self.connack = None

        self.disconnected = False
        self.connected = False

    def __str__(self):
        return str(self.messages) + str(self.messagedicts) + str(self.publisheds) + \
               str(self.subscribeds) + str(self.unsubscribeds) + str(self.disconnected)

    def on_message(self, client, topic, payload, qos, properties):
        super().on_message(client, topic, payload, qos, properties)
        logging.info('[RECV MSG {}] TOPIC: {} PAYLOAD: {} QOS: {} PROPERTIES: {}'
                     .format(client._client_id, topic, payload, qos, properties))
        self.messages.append((topic, payload, qos, properties))

    def on_subscribe(self, client, mid, qos):
        super().on_subscribe(client, mid, qos)
        logging.info('[SUBSCRIBED {}] QOS: {}'.format(client._client_id, qos))
        self.subscribeds.append(mid)


async def clean_retained(host, port, username, password=None):
    def on_message(client, topic, payload, qos, properties):
        curclient.publish(topic, b"", qos=0, retain=True)

    curclient = gmqtt.Client("clean retained".encode("utf-8"), clean_session=True)

    curclient.set_auth_credentials(username, password)
    curclient.on_message = on_message
    await curclient.connect(host=host, port=port)
    curclient.subscribe("#")
    await asyncio.sleep(10)  # wait for all retained messages to arrive
    await curclient.disconnect()
    time.sleep(.1)


async def cleanup(host, port=1883, username=None, password=None, client_ids=None):
    # clean all client state
    print("clean up starting")
    client_ids = client_ids or ("myclientid", "myclientid2", "myclientid3")

    for clientid in client_ids:
        curclient = gmqtt.Client(clientid.encode("utf-8"), clean_session=True)
        curclient.set_auth_credentials(username=username, password=password)
        await curclient.connect(host=host, port=port)
        time.sleep(.1)
        await curclient.disconnect()
        time.sleep(.1)

    # clean retained messages
    await clean_retained(host, port, username, password=password)

